require_relative "../util/Comp"

module Callidus
    class ExponentialRegression
        attr_accessor :input
        attr_accessor :output
        attr_accessor :predicted_output

        attr_reader :a
        attr_reader :b

        attr_reader :correlation
        attr_reader :standard_error

        def initialize(ip = [], op = [])
            @input = ip
            @output = op

            @trained = false
        end

        private def assert_trained(method = "train")
            if !@trained
                raise Util::UntrainedError.new("ExponentialRegression")
            end
        end

        def formatted
            assert_trained()

            "f(x) = #{@a}e^#{@b}x"
        end

        def train
            mean_x = @input.mean
            mean_lny = @output.inject(0.0) { |s, n|  s + Math.log(n) } / @output.size

            sum_xx = 0.0
            sum_xy = 0.0
            sum_yy = 0.0

            @input.each_index do |i|
                sum_xx += @input[i] * @input[i]
                sum_xy += @input[i] * Math.log(@output[i])
                sum_yy += Math.log(@output[i]) * Math.log(@output[i])
            end

            sum_xx = (sum_xx / @input.size) - (mean_x * mean_x)
            sum_xy = (sum_xy / @input.size) - (mean_x * mean_lny)
            sum_yy = (sum_yy / @output.size) - (mean_lny * mean_lny)

            @b = sum_xy / sum_xx
            @a = 2.7182818284590452353602 ** (mean_lny - b * mean_x)

            @predicted_output = @input.map { |x| (@a * (2.7182818284590452353602 ** (@b * x))).round(3) }

            @correlation = (sum_xy / (Math.sqrt(sum_xx) * Math.sqrt(sum_yy))).round(3)

            @trained = true

            @a = @a.round(3)
            @b = @b.round(3)

            self.find_standard_error
        end

        def find_standard_error
            assert_trained()

            n = @input.size > @output.size ? @output.size : @input.size

            diffs = []

            n.times do |i|
                diffs << (@predicted_output[i] - @output[i]) ** 2
            end

            @standard_error = (Math.sqrt(diffs.sum/(n - 2.0))).round(3)

            self
        end

        def predict(x)
            assert_trained()

            @a * (2.7182818284590452353602 ** (@b * x))
        end
    end
end