require_relative "../util/Util"

module Callidus
    class LinearRegression
        attr_accessor :input
        attr_accessor :output
        attr_accessor :predicted_output

        attr_reader :slope
        attr_reader :y_intercept

        attr_reader :correlation
        attr_reader :standard_error

        def initialize(ip = [], op = [])
            @input = ip
            @output = op

            @trained = false
        end

        private def assert_trained(method = "train")
            if !@trained
                raise Util::UntrainedError.new("LinearRegression")
            end
        end

        def formatted
            assert_trained()

            "f(x) = #{@slope}x + #{@y_intercept}"
        end

        def train
            sum_x = 0
            sum_y = 0
            sum_xy = 0
            sum_xx = 0

            n = @input.size > @output.size ? @output.size : @input.size

            @predicted_output = []

            n.times do |i|
                x = @input[i]
                y = @output[i]

                sum_x += x
                sum_y += y
                sum_xx += (x * x)
                sum_xy += (x * y)
            end

            @slope = ((n * sum_xy - sum_x * sum_y).to_f / (n * sum_xx - sum_x * sum_x).to_f).round(3)
            @y_intercept = ((sum_y / n).to_f - (@slope * sum_x).to_f / n.to_f).round(3)

            @predicted_output = @input.map { |x| (x * @slope + @y_intercept).round(3) }

            @trained = true

            self.find_correlation.find_standard_error
        end

        def find_correlation
            assert_trained()

            mean = @output.sum/@output.size

            diffYM = @output.map { |y| (y - mean) ** 2 };
            diffPYM = @predicted_output.map { |y| (y - mean) ** 2 };

            @correlation = (diffPYM.sum/diffYM.sum).round(3)

            self
        end

        def find_standard_error
            assert_trained()

            n = @input.size > @output.size ? @output.size : @input.size

            diffs = []

            n.times do |i|
                diffs << (@predicted_output[i] - @output[i]) ** 2
            end

            @standard_error = (Math.sqrt(diffs.sum/(n - 2.0))).round(3)

            self
        end

        def predict(x)
            assert_trained()
            
            x * @slope + @y_intercept
        end
    end
end
