require_relative "../util/Util"

module Callidus
    class Winnow
        attr_accessor :input
        attr_accessor :output

        attr_accessor :a
        attr_accessor :mode

        attr_reader :weights

        def initialize(ip = [], op = [], options = {})
            @input = ip
            @output = op

            @default_weight = options[:default_weight] || 1
            @a = options[:a] || 2;
            @mode = options[:mode] || 0 # 0 is demote, 1 is reset

            @weights = []
            @trained = false
        end

        private def assert_trained(method = "train")
            if !@trained
                raise Util::UntrainedError.new("Winnow")
            end
        end

        def train(a = @a)
            if (@input.size > 0 and @input[0] and @input[0].size > 0 and @input[0].size != @weights.size)
                @weights = Array.new(@input[0].size, @default_weight)
            end
    
            @input.each_index do |i|
                @input[i].each_index do |j|
                    if @input[i][j] == 1
                        @weights[j] *= @output[i] == 0 ? (@mode == 0 ? 1 / a : 0) : a
                    end
                end
            end
    
            @trained = true
    
            self
        end

        def predict(x, options = {})
            assert_trained()

            bias = options[:bias] || 0
            threshold = options[:threshold] || (@input.size / 2).to_i
    
            probability = 0;
    
            x.each_index do |i|
                probability += @weights[i] * x[i] + bias
            end
    
            (probability > threshold ? 1 : 0)
        end
    end
end
